#define MOZ_SOURCE_STAMP 865939f4946d80fa532aaa168515a2fe69f9a774
#define MOZ_SOURCE_REPO https://hg.mozilla.org/releases/mozilla-beta
#define MOZ_SOURCE_URL https://hg.mozilla.org/releases/mozilla-beta/rev/865939f4946d80fa532aaa168515a2fe69f9a774
